#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Reset noise models."""
import typing as t

import astropy.constants as const
import numpy as np

from pyxel.detectors import Detector
from pyxel.util import temporary_random_state


def compute_ktc_noise(
    temperature: float, capacitance: float, shape: t.Tuple[int, int]
) -> np.ndarray:
    """Compute KTC noise array.

    Parameters
    ----------
    temperature: float
        Temperature. Unit: K
    capacitance: float
        Node capacitance. Unit: F
    shape: tuple
        Shape of the output array.

    Returns
    -------
    np.ndarray
    """

    rms = np.sqrt(const.k_B.value * temperature / capacitance)

    return np.random.normal(0, rms, shape)


@temporary_random_state
def ktc_noise(
    detector: Detector,
    node_capacitance: t.Optional[float] = None,
    seed: t.Optional[int] = None,
) -> None:
    """Apply KTC reset noise to detector signal array.

    Parameters
    ----------
    detector : Detector
        Pyxel detector object.
    node_capacitance: float, optional
        Node capacitance. Unit: F
    seed: int, optional
        Random seed.
    """

    if node_capacitance:

        if node_capacitance <= 0:
            raise ValueError("Node capacitance should be larger than 0!")

        detector.signal.array += compute_ktc_noise(
            temperature=detector.environment.temperature,
            capacitance=node_capacitance,
            shape=detector.geometry.shape,
        )

    else:
        try:
            capacitance = detector.characteristics.node_capacitance

            detector.signal.array += compute_ktc_noise(
                temperature=detector.environment.temperature,
                capacitance=capacitance,
                shape=detector.geometry.shape,
            )

        except AttributeError as ex:
            raise AttributeError(
                "Characteristic node_capacitance not available for the detector used. "
                "Please specify node_capacitance in the model argument!"
            ) from ex



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

""":term:`MKID`-array detector modeling class."""

import typing as t

from pyxel.data_structure import Phase
from pyxel.detectors import Detector
from pyxel.util.memory import memory_usage_details

if t.TYPE_CHECKING:
    import pandas as pd

    from pyxel.detectors import Environment, MKIDCharacteristics, MKIDGeometry


class MKID(Detector):
    """:term:`MKID`-based detector class containing all detector attributes and data."""

    def __init__(
        self,
        geometry: "MKIDGeometry",
        environment: "Environment",
        characteristics: "MKIDCharacteristics",
    ):
        self._geometry = geometry  # type: MKIDGeometry
        self._characteristics = characteristics  # type: MKIDCharacteristics
        self._phase = None  # type: t.Optional[Phase]

        super().__init__(environment=environment)
        self.reset()

    def __eq__(self, other) -> bool:
        return (
            type(self) == type(other)
            and self.geometry == other.geometry
            and self.environment == other.environment
            and self.characteristics == other.characteristics
            and self._phase == other._phase
            and super().__eq__(other)
        )

    def reset(self) -> None:
        """TBW."""
        super().reset()
        self._phase = Phase(geo=self.geometry)

    def empty(self, empty_all: bool = True) -> None:
        """Empty the data in the detector.

        Returns
        -------
        None
        """
        super().empty(empty_all)

        if empty_all and self._phase:
            self.phase.array *= 0

    @property
    def geometry(self) -> "MKIDGeometry":
        """TBW."""
        return self._geometry

    @property
    def characteristics(self) -> "MKIDCharacteristics":
        """TBW."""
        return self._characteristics

    @property
    def phase(self) -> Phase:
        """TBW."""
        if not self._phase:
            raise RuntimeError("'phase' not initialized.")

        return self._phase

    def memory_usage(
        self, print_result: bool = True, human_readable: bool = True
    ) -> dict:
        """TBW.

        Returns
        -------
        dict
            Dictionary of attribute memory usage
        """
        attributes = [
            "_photon",
            "_charge",
            "_pixel",
            "_signal",
            "_image",
            "_phase",
            "material",
            "environment",
            "_geometry",
            "_characteristics",
        ]

        return memory_usage_details(
            self, attributes, print_result=print_result, human_readable=human_readable
        )

    # TODO: Refactor this
    def to_dict(self) -> t.Mapping:
        """Convert an instance of `MKID` to a `dict`."""
        dct = {
            "version": 1,
            "type": "MKID",
            "properties": {
                "geometry": self.geometry.to_dict(),
                "environment": self.environment.to_dict(),
                "characteristics": self.characteristics.to_dict(),
            },
            "data": {
                "photon": None if self._photon is None else self._photon.array.copy(),
                "pixel": None if self._pixel is None else self._pixel.array.copy(),
                "signal": None if self._signal is None else self._signal.array.copy(),
                "image": None if self._image is None else self._image.array.copy(),
                "phase": None if self._phase is None else self._phase.array.copy(),
                "charge": None
                if self._charge is None
                else {
                    "array": self._charge.array.copy(),
                    "frame": self._charge.frame.copy(),
                },
            },
        }

        return dct

    # TODO: Refactor this
    @classmethod
    def from_dict(cls, dct: t.Mapping) -> "MKID":
        """Create a new instance of `MKID` from a `dict`."""
        # TODO: This is a simplistic implementation. Improve this.
        from pyxel.detectors import Environment, MKIDCharacteristics, MKIDGeometry

        if dct["type"] != "MKID":
            raise ValueError

        if dct["version"] != 1:
            raise ValueError

        properties = dct["properties"]
        geometry = MKIDGeometry.from_dict(properties["geometry"])
        environment = Environment.from_dict(properties["environment"])
        characteristics = MKIDCharacteristics.from_dict(properties["characteristics"])

        detector = cls(
            geometry=geometry,
            environment=environment,
            characteristics=characteristics,
        )

        data = dct["data"]

        if "photon" in data:
            detector.photon.array = data["photon"]
        if "pixel" in data:
            detector.pixel.array = data["pixel"]
        if "signal" in data:
            detector.signal.array = data["signal"]
        if "image" in data:
            detector.image.array = data["image"]
        if "phase" in data and data["phase"] is not None:
            detector.phase.array = data["phase"]
        if "charge" in data and data["charge"] is not None:
            charge_dct = data["charge"]
            detector.charge._array = charge_dct["array"]

            new_frame = charge_dct["frame"]  # type: pd.DataFrame
            previous_frame = detector.charge._frame  # type: pd.DataFrame
            detector.charge._frame = new_frame[previous_frame.columns]

        return detector



#   Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#  #
#   This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#   is part of this Pyxel package. No part of the package, including
#   this file, may be copied, modified, propagated, or distributed except according to
#   the terms contained in the file ‘LICENCE.txt’.

"""Models to generate charge due to dark current process."""

import typing as t

import numpy as np
from astropy import constants as const

from pyxel.detectors import APD, Detector
from pyxel.util import temporary_random_state


def band_gap(band_gap_0: float, alpha: float, beta: float, temperature: float) -> float:
    """Return band gap based on Varshni empirical expression.

    Parameters
    ----------
    band_gap_0: float
        Parameter E_0. Unit: eV
    alpha: float
        Alpha parameter. Unit: eV/K
    beta: float
        Beta parameter. Unit: K
    temperature:
        Temperature. Unit K

    Returns
    -------
    gap: float
        Band gap value. Unit: eV
    """

    gap = band_gap_0 - (alpha * temperature**2) / (temperature + beta)

    return gap


def band_gap_silicon(temperature: float) -> float:
    """Return band gap in Silicon based on Varshni empirical expression.

    Parameters
    ----------
    temperature: float
        Temperature. Unit: K

    Returns
    -------
    float
        Band gap value. Unit: eV
    """

    band_gap_0 = 1.1557  # eV
    alpha = 7.021e-4  # ev/K
    beta = 1108.0  # K

    return band_gap(
        band_gap_0=band_gap_0, alpha=alpha, beta=beta, temperature=temperature
    )


def average_dark_current(
    temperature: float,
    pixel_area: float,
    figure_of_merit: float,
    band_gap: float,
    band_gap_room_temperature: float,
) -> float:
    """Compute average dark current.

    Parameters
    ----------
    temperature: float
        Temperature. Unit: K
    pixel_area:
        Pixel area. Unit: cm^2
    figure_of_merit: float
        Dark current figure of merit. Unit: nA/cm^2
    band_gap: float
        Semiconductor band_gap. Unit: eV
    band_gap_room_temperature: float
        Semiconductor band gap at 300K. If none, the one for silicon is used. Unit: eV

    Returns
    -------
    avg_dark_current: float
        Average dark current. Unit: e-/pixel/s
    """

    k_b = const.k_B.value
    e_0 = const.e.value

    room_temperature = 300

    room_temperature_factor = room_temperature ** (3 / 2) * np.exp(
        -band_gap_room_temperature * e_0 / (2 * k_b * room_temperature)
    )

    avg_dark_current = (
        pixel_area  # in cm^2
        * figure_of_merit  # in nA/cm^2
        * 1e-9  # conversion to A/cm^2
        * (1 / e_0)  # conversion to e-/s/cm^2
        * temperature ** (3 / 2)
        * np.exp(-band_gap * e_0 / (2 * k_b * temperature))
        * (1 / room_temperature_factor)
    )  # Unit: e-/s/pixel

    return avg_dark_current


def compute_dark_current(
    shape: t.Tuple[int, int],
    time_step: float,
    temperature: float,
    pixel_area: float,
    figure_of_merit: float,
    band_gap: float,
    fixed_pattern_noise_factor: float,
    band_gap_room_temperature: float,
) -> np.ndarray:
    """Compute dark current.

    Based on:
    Konnik, Mikhail V. and James S. Welsh.
    “High-level numerical simulations of noise in CCD and CMOS photosensors: review and tutorial.”
    ArXiv abs/1412.4031 (2014): n. pag.

    Parameters
    ----------
    shape: tuple
        Output array shape.
    time_step: float
        Time step. Unit: s
    temperature: float
        Temperature. Unit: K
    pixel_area:
        Pixel area. Unit: cm^2
    figure_of_merit: float
        Dark current figure of merit. Unit: nA/cm^2
    band_gap: float
        Semiconductor band_gap. Unit: eV
    fixed_pattern_noise_factor: float
        Fixed pattern noise factor.
    band_gap_room_temperature: float
        Semiconductor band gap at 300K. If none, the one for silicon is used. Unit: eV

    Returns
    -------
    dark_current_2d: ndarray
        Dark current values. Unit: e-
    """
    avg_dark_current = average_dark_current(
        temperature=temperature,
        pixel_area=pixel_area,
        figure_of_merit=figure_of_merit,
        band_gap=band_gap,
        band_gap_room_temperature=band_gap_room_temperature,
    )

    dark_signal_2d = np.ones(shape) * avg_dark_current * time_step
    dark_current_shot_noise_2d = np.random.poisson(
        dark_signal_2d
    )  # dark current shot noise
    dark_current_fpn_sigma = (
        time_step * avg_dark_current * fixed_pattern_noise_factor
    )  # sigma of fpn distribution

    dark_current_2d = dark_current_shot_noise_2d * (
        1 + np.random.lognormal(sigma=dark_current_fpn_sigma, size=shape)
    )

    return dark_current_2d


@temporary_random_state
def dark_current(
    detector: Detector,
    figure_of_merit: float,
    fixed_pattern_noise_factor: float,
    band_gap: t.Optional[float] = None,
    band_gap_room_temperature: t.Optional[float] = None,
    seed: t.Optional[int] = None,
) -> None:
    """Add dark current to the detector charge.

    Based on:
    Konnik, Mikhail V. and James S. Welsh.
    “High-level numerical simulations of noise in CCD and CMOS photosensors: review and tutorial.”
    ArXiv abs/1412.4031 (2014): n. pag.

    Parameters
    ----------
    detector: Detector
        Pyxel detector object.
    figure_of_merit: float
        Dark current figure of merit. Unit: nA/cm^2
    fixed_pattern_noise_factor: float
        Fixed pattern noise factor.
    band_gap: float, optional
        Semiconductor band_gap. If none, the one for silicon is used. Unit: eV
    band_gap_room_temperature: float, optional
        Semiconductor band gap at 300K. If none, the one for silicon is used. Unit: eV
    seed: int, optional
        Random seed.
    """
    geo = detector.geometry
    pixel_area = geo.pixel_vert_size * 1e-4 * geo.pixel_horz_size * 1e-4  # in cm^2
    temperature = detector.environment.temperature
    time_step = detector.time_step

    if band_gap and band_gap_room_temperature:
        final_band_gap = band_gap
        final_band_gap_room_temperature = band_gap_room_temperature
    elif band_gap or band_gap_room_temperature:
        raise ValueError(
            "Both parameters band_gap and band_gap_room_temperature have to be defined."
        )
    else:
        final_band_gap = band_gap_silicon(temperature)
        final_band_gap_room_temperature = band_gap_silicon(temperature=300)

    dark_current_array = compute_dark_current(
        shape=geo.shape,
        time_step=time_step,
        temperature=temperature,
        pixel_area=pixel_area,
        figure_of_merit=figure_of_merit,
        band_gap=final_band_gap,
        band_gap_room_temperature=final_band_gap_room_temperature,
        fixed_pattern_noise_factor=fixed_pattern_noise_factor,
    )

    detector.charge.add_charge_array(dark_current_array)


def calculate_simple_dark_current(
    num_rows: int, num_cols: int, current: float, exposure_time: float
) -> np.ndarray:
    """Simulate dark current in a :term:`CCD`.

    Parameters
    ----------
    num_rows : int
        Number of rows for the generated image.
    num_cols : int
        Number of columns for the generated image.
    current : float
        Dark current, in electrons/pixel/second
    exposure_time : float
        Length of the simulated exposure, in seconds.

    Returns
    -------
    dark_im_array_2d: ndarray
        An array the same shape and dtype as the input containing dark counts
        in units of charge (e-).
    """
    # dark current for every pixel
    mean_dark_charge = current * exposure_time

    # This random number generation should change on each call.
    dark_im_array_2d = np.random.poisson(mean_dark_charge, size=(num_rows, num_cols))
    return dark_im_array_2d


@temporary_random_state
def simple_dark_current(
    detector: Detector, dark_rate: float, seed: t.Optional[int] = None
) -> None:
    """Simulate dark current in a detector.

    Parameters
    ----------
    detector : Detector
        Any detector object.
    dark_rate : float
        Dark current, in electrons/pixel/second, which is the way
        manufacturers typically report it.
    seed: int, optional
    """

    exposure_time = detector.time_step
    geo = detector.geometry

    dark_current_array = calculate_simple_dark_current(
        num_rows=geo.row,
        num_cols=geo.col,
        current=dark_rate,
        exposure_time=exposure_time,
    ).astype(
        float
    )  # type: np.ndarray

    detector.charge.add_charge_array(dark_current_array)


def calculate_dark_current_saphira(
    temperature: float,
    avalanche_gain: float,
    shape: t.Tuple[int, int],
    exposure_time: float,
) -> np.ndarray:
    """Simulate dark current in a Saphira :term:`APD`.

    From: I. M. Baker et al., Linear-mode avalanche photodiode arrays in HgCdTe at Leonardo, UK: the
    current status, in Image Sensing Technologies: Materials, Devices, Systems, and Applications VI,
    2019, vol. 10980, no. May, p. 20.

    Parameters
    ----------
    temperature
    avalanche_gain
    shape
    exposure_time : float
        Length of the simulated exposure, in seconds.

    Returns
    -------
    dark_im_array_2d: ndarray
        An array the same shape and dtype as the input containing dark counts
        in units of charge (e-).
    """

    # We can split the dark current vs. gain vs. temp plot ([5] Fig. 3) into three linear
    # 'regimes': 1) low-gain, low dark current; 2) nominal; and 3) trap-assisted tunneling.
    # The following ignores (1) for now since this only applies at gains less than ~2.

    dark_nominal = (15.6e-12 * np.exp(0.2996 * temperature)) * (
        avalanche_gain
        ** ((-5.43e-4 * (temperature**2)) + (0.0669 * temperature) - 1.1577)
    )
    dark_tunnel = 3e-5 * (avalanche_gain**3.2896)

    # The value we want is the maximum of the two regimes (see the plot):
    dark = max(dark_nominal, dark_tunnel)

    # dark current for every pixel
    mean_dark_charge = dark * exposure_time

    # This random number generation should change on each call.
    dark_im_array_2d = np.random.poisson(mean_dark_charge, size=shape)

    return dark_im_array_2d


@temporary_random_state
def dark_current_saphira(detector: APD, seed: t.Optional[int] = None) -> None:
    """Simulate dark current in a Saphira APD detector.

    Reference: I. M. Baker et al., Linear-mode avalanche photodiode arrays in HgCdTe at Leonardo, UK: the
    current status, in Image Sensing Technologies: Materials, Devices, Systems, and Applications VI,
    2019, vol. 10980, no. May, p. 20.

    Parameters
    ----------
    detector : APD
        An APD detector object.
    seed: int, optional
    """

    if not isinstance(detector, APD):
        raise TypeError("Expecting an APD object for detector.")
    if detector.environment.temperature > 100:
        raise ValueError(
            "Dark current estimation is inaccurate for temperatures more than 100 K!"
        )
    if detector.characteristics.avalanche_gain < 2:
        raise ValueError("Dark current is inaccurate for avalanche gains less than 2!")

    exposure_time = detector.time_step

    dark_current_array = calculate_dark_current_saphira(
        temperature=detector.environment.temperature,
        avalanche_gain=detector.characteristics.avalanche_gain,
        shape=detector.geometry.shape,
        exposure_time=exposure_time,
    ).astype(
        float
    )  # type: np.ndarray

    detector.charge.add_charge_array(dark_current_array)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""CCD detector modeling class."""

import typing as t

from pyxel.detectors import Detector

if t.TYPE_CHECKING:
    import pandas as pd

    from pyxel.detectors import CCDCharacteristics, CCDGeometry, Environment


class CCD(Detector):
    """Charge-Coupled Device class containing all detector attributes and data."""

    def __init__(
        self,
        geometry: "CCDGeometry",
        environment: "Environment",
        characteristics: "CCDCharacteristics",
    ):
        self._geometry = geometry  # type: CCDGeometry
        self._characteristics = characteristics  # type: CCDCharacteristics

        super().__init__(environment=environment)
        super().reset()

    def __eq__(self, other) -> bool:
        return (
            type(self) == type(other)
            and self.geometry == other.geometry
            and self.environment == other.environment
            and self.characteristics == other.characteristics
            and super().__eq__(other)
        )

    @property
    def geometry(self) -> "CCDGeometry":
        """TBW."""
        return self._geometry

    @property
    def characteristics(self) -> "CCDCharacteristics":
        """TBW."""
        return self._characteristics

    # TODO: Refactor this
    def to_dict(self) -> t.Mapping:
        """Convert an instance of `CCD` to a `dict`."""
        dct = {
            "version": 1,
            "type": "CCD",
            "properties": {
                "geometry": self.geometry.to_dict(),
                "environment": self.environment.to_dict(),
                "characteristics": self.characteristics.to_dict(),
            },
            "data": {
                "photon": None if self._photon is None else self._photon.array.copy(),
                "pixel": None if self._pixel is None else self._pixel.array.copy(),
                "signal": None if self._signal is None else self._signal.array.copy(),
                "image": None if self._image is None else self._image.array.copy(),
                "charge": None
                if self._charge is None
                else {
                    "array": self._charge.array.copy(),
                    "frame": self._charge.frame.copy(),
                },
            },
        }

        return dct

    # TODO: Refactor this
    @classmethod
    def from_dict(cls, dct: t.Mapping) -> "CCD":
        """Create a new instance of `CCD` from a `dict`."""
        # TODO: This is a simplistic implementation. Improve this.
        from pyxel.detectors import CCDCharacteristics, CCDGeometry, Environment

        if dct["type"] != "CCD":
            raise ValueError

        if dct["version"] != 1:
            raise ValueError

        properties = dct["properties"]
        geometry = CCDGeometry.from_dict(properties["geometry"])
        environment = Environment.from_dict(properties["environment"])
        characteristics = CCDCharacteristics.from_dict(properties["characteristics"])

        detector = cls(
            geometry=geometry,
            environment=environment,
            characteristics=characteristics,
        )

        data = dct["data"]

        if "photon" in data:
            detector.photon.array = data["photon"]
        if "pixel" in data:
            detector.pixel.array = data["pixel"]
        if "signal" in data:
            detector.signal.array = data["signal"]
        if "image" in data:
            detector.image.array = data["image"]
        if "charge" in data and data["charge"] is not None:
            charge_dct = data["charge"]
            detector.charge._array = charge_dct["array"]

            new_frame = charge_dct["frame"]  # type: pd.DataFrame
            previous_frame = detector.charge._frame  # type: pd.DataFrame
            detector.charge._frame = new_frame[previous_frame.columns]

        return detector



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Voltage offset models."""
import typing as t

import numpy as np

from pyxel.detectors import APD, Detector


def compute_dc_offset(offset: float, shape: t.Tuple[int, int]) -> np.ndarray:
    """Compute DC offset voltage array.

    Parameters
    ----------
    offset: float
        DC voltage offset. Unit: V
    shape: tuple
        Shape of the output array.

    Returns
    -------
    np.ndarray
    """
    return np.ones(shape) * offset


def dc_offset(detector: Detector, offset: float) -> None:
    """Apply DC voltage to the detector.

    Parameters
    ----------
    detector: Detector
        Pyxel Detector object.
    offset: float
        DC offset voltage. Unit: V
    """

    if (
        not min(detector.characteristics.adc_voltage_range)
        < offset
        < max(detector.characteristics.adc_voltage_range)
    ):
        raise ValueError("Parameter 'offset' out of bonds of the ADC voltage range.")

    detector.signal.array += compute_dc_offset(offset, detector.geometry.shape)


def compute_output_pixel_reset_voltage_apd(
    roic_drop: float,
    roic_gain: float,
    pixel_reset_voltage: float,
    shape: t.Tuple[int, int],
) -> np.ndarray:
    """Compute output pixel reset voltage.

    Parameters
    ----------
    roic_drop: float
        Readout circuit drop voltage. Unit: V
    roic_gain: float
        Readout circuit gain.
    pixel_reset_voltage:
        Pixel reset voltage. Unit: V
    shape: tuple
        Shape of output array.

    Returns
    -------
    np.ndarray
    """

    value = (roic_gain * pixel_reset_voltage) - (
        roic_drop * roic_gain
    )  # Reproduces linear plot of PRV vs. output voltage

    return np.ones(shape) * value


def output_pixel_reset_voltage_apd(detector: APD, roic_drop: float) -> None:
    """Apply output pixel reset voltage to APD detector.

    Parameters
    ----------
    detector: APD
        Pyxel APD object.
    roic_drop: float
        Readout circuit drop voltage. Unit: V
    """
    if not isinstance(detector, APD):
        raise TypeError("Expecting a 'APD' detector object.")

    ch = detector.characteristics
    shape = detector.geometry.shape

    offset = compute_output_pixel_reset_voltage_apd(
        roic_drop=roic_drop,
        roic_gain=ch.roic_gain,
        pixel_reset_voltage=ch.pixel_reset_voltage,
        shape=shape,
    )

    if (
        not min(detector.characteristics.adc_voltage_range)
        < offset[0, 0]
        < max(detector.characteristics.adc_voltage_range)
    ):
        raise ValueError(
            "Output pixel reset voltage out of bonds of the ADC voltage range."
        )

    detector.signal.array += max(ch.adc_voltage_range) - offset



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""TBW.

References
----------
[1] Leonardo MW Ltd., Electrical Interface Document for SAPHIRA (ME1000) in a 68-pin LLC
(3313-520-), issue 1A, 2012.

[2] I. Pastrana et al., HgCdTe SAPHIRA arrays: individual pixel measurement of charge gain and
node capacitance utilizing a stable IR LED, in High Energy, Optical, and Infrared Detectors for
Astronomy VIII, 2018, vol. 10709, no. July 2018, 2018, p. 37.

[3] S. B. Goebel et al., Overview of the SAPHIRA detector for adaptive optics applications, in
Journal of Astronomical Telescopes, Instruments, and Systems, 2018, vol. 4, no. 02, p. 1.

[4] G. Finger et al., Sub-electron read noise and millisecond full-frame readout with the near
infrared eAPD array SAPHIRA, in Adaptive Optics Systems V, 2016, vol. 9909, no. July 2016, p.
990912.

[5] I. M. Baker et al., Linear-mode avalanche photodiode arrays in HgCdTe at Leonardo, UK: the
current status, in Image Sensing Technologies: Materials, Devices, Systems, and Applications VI,
2019, vol. 10980, no. May, p. 20.
"""


import math
import typing as t

import numpy as np
from toolz import dicttoolz

from pyxel.util.memory import get_size


class APDCharacteristics:
    """Characteristic attributes of the APD detector.

    Parameters
    ----------
    roic_gain:
        Gain of the read-out integrated circuit. Unit: V/V
    quantum_efficiency: float, optional
        Quantum efficiency.
    full_well_capacity: float, optional
        Full well capacity. Unit: e-
    adc_bit_resolution: int, optional
        ADC bit resolution.
    adc_voltage_range: tuple of floats, optional
        ADC voltage range. Unit: V
    avalanche_gain: float, optional
        APD gain. Unit: electron/electron
    pixel_reset_voltage: float
        DC voltage going into the detector, not the voltage of a reset pixel. Unit: V
    common_voltage: float
        Common voltage. Unit: V
    """

    def __init__(
        self,
        roic_gain: float,  # unit: V
        quantum_efficiency: t.Optional[float] = None,  # unit: NA
        full_well_capacity: t.Optional[float] = None,  # unit: electron
        adc_bit_resolution: t.Optional[int] = None,
        adc_voltage_range: t.Optional[t.Tuple[float, float]] = None,  # unit: V
        avalanche_gain: t.Optional[float] = None,  # unit: electron/electron
        pixel_reset_voltage: t.Optional[float] = None,  # unit: V
        common_voltage: t.Optional[float] = None,  # unit: V
    ):

        self._avalanche_gain = avalanche_gain
        self._common_voltage = common_voltage
        self._pixel_reset_voltage = pixel_reset_voltage

        if avalanche_gain and pixel_reset_voltage and common_voltage:
            raise ValueError(
                "Please only specify two inputs out of: avalanche gain, pixel reset voltage, common voltage."
            )
        elif avalanche_gain and pixel_reset_voltage and not common_voltage:
            self._avalanche_bias = self.gain_to_bias_saphira(avalanche_gain)
            self._common_voltage = pixel_reset_voltage - self.avalanche_bias
        elif avalanche_gain and common_voltage and not pixel_reset_voltage:
            self._avalanche_bias = self.gain_to_bias_saphira(avalanche_gain)
            self._pixel_reset_voltage = common_voltage + self.avalanche_bias
        elif common_voltage and pixel_reset_voltage and not avalanche_gain:
            self._avalanche_bias = pixel_reset_voltage - common_voltage
            self._avalanche_gain = self.bias_to_gain_saphira(self.avalanche_bias)
        else:
            raise ValueError(
                "Not enough input parameters provided to calculate avalanche bias!"
            )

        if quantum_efficiency and not (0.0 <= quantum_efficiency <= 1.0):
            raise ValueError("'quantum_efficiency' must be between 0.0 and 1.0.")
        if avalanche_gain and not (1.0 <= avalanche_gain <= 1000.0):
            raise ValueError("'apd_gain' must be between 1.0 and 1000.0.")
        if adc_bit_resolution and not (4 <= adc_bit_resolution <= 64):
            raise ValueError("'adc_bit_resolution' must be between 4 and 64.")
        if adc_voltage_range and not len(adc_voltage_range) == 2:
            raise ValueError("Voltage range must have length of 2.")
        if full_well_capacity and not (0.0 <= full_well_capacity <= 1.0e7):
            raise ValueError("'full_well_capacity' must be between 0 and 1e7.")

        self._quantum_efficiency = quantum_efficiency
        self._full_well_capacity = full_well_capacity
        self._adc_voltage_range = adc_voltage_range
        self._adc_bit_resolution = adc_bit_resolution
        self._node_capacitance = self.bias_to_node_capacitance_saphira(
            self.avalanche_bias
        )  # type: float
        self._roic_gain = roic_gain
        self._charge_to_volt_conversion = self.detector_gain_saphira(
            capacitance=self.node_capacitance, roic_gain=self.roic_gain
        )  # type: float
        self._numbytes = 0

    def __eq__(self, other) -> bool:
        return (
            type(self) == type(other)
            and self._quantum_efficiency == other._quantum_efficiency
            and self._full_well_capacity == other._full_well_capacity
            and self._adc_bit_resolution == other._adc_bit_resolution
            and self._adc_voltage_range == other._adc_voltage_range
            and self._avalanche_gain == other._avalanche_gain
            and self._pixel_reset_voltage == other._pixel_reset_voltage
            and self._common_voltage == other._common_voltage
        )

    @property
    def quantum_efficiency(self) -> float:
        """Get Quantum efficiency."""
        if self._quantum_efficiency:
            return self._quantum_efficiency
        else:
            raise ValueError(
                "'quantum_efficiency' not specified in detector characteristics."
            )

    @quantum_efficiency.setter
    def quantum_efficiency(self, value: float) -> None:
        """Set Quantum efficiency."""
        if np.min(value) < 0.0 and np.max(value) <= 1.0:
            raise ValueError("'quantum_efficiency' values must be between 0.0 and 1.0.")

        self._quantum_efficiency = value

    @property
    def avalanche_gain(self) -> float:
        """Get APD gain."""
        if self._avalanche_gain:
            return self._avalanche_gain
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @avalanche_gain.setter
    def avalanche_gain(self, value: float) -> None:
        """Set APD gain."""
        if np.min(value) < 1.0 and np.max(value) <= 1000.0:
            raise ValueError("'apd_gain' values must be between 1.0 and 1000.")
        self._avalanche_gain = value
        self._avalanche_bias = self.gain_to_bias_saphira(value)
        self._common_voltage = self.pixel_reset_voltage - self.avalanche_bias

    @property
    def pixel_reset_voltage(self) -> float:
        """Get pixel reset voltage."""
        if self._pixel_reset_voltage:
            return self._pixel_reset_voltage
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @pixel_reset_voltage.setter
    def pixel_reset_voltage(self, value: float) -> None:
        """Set pixel reset voltage."""
        self._avalanche_bias = value - self.common_voltage
        self._avalanche_gain = self.bias_to_gain_saphira(self.avalanche_bias)
        self._pixel_reset_voltage = value

    @property
    def common_voltage(self) -> float:
        """Get common voltage."""
        if self._common_voltage:
            return self._common_voltage
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @common_voltage.setter
    def common_voltage(self, value: float) -> None:
        """Set common voltage."""
        self._avalanche_bias = self.pixel_reset_voltage - value
        self._avalanche_gain = self.bias_to_gain_saphira(self.avalanche_bias)
        self._common_voltage = value

    @property
    def avalanche_bias(self) -> float:
        """Get avalanche bias."""
        if self._avalanche_bias:
            return self._avalanche_bias
        else:
            raise ValueError("'apd_gain' not specified in detector characteristics.")

    @property
    def roic_gain(self) -> float:
        """Get roic gainn."""
        if self._roic_gain:
            return self._roic_gain
        else:
            raise ValueError("'roic_gain' not specified in detector characteristics.")

    @property
    def node_capacitance(self) -> float:
        """Get node capacitance."""
        self._node_capacitance = self.bias_to_node_capacitance_saphira(
            self.avalanche_bias
        )
        return self._node_capacitance

    @property
    def charge_to_volt_conversion(self):
        """Get charge to voltage conversion factor."""
        self._charge_to_volt_conversion = self.detector_gain_saphira(
            capacitance=self.node_capacitance, roic_gain=self.roic_gain
        )
        return self._charge_to_volt_conversion

    @property
    def adc_bit_resolution(self) -> int:
        """Get bit resolution of the Analog-Digital Converter."""
        if self._adc_bit_resolution:
            return self._adc_bit_resolution
        else:
            raise ValueError(
                "'adc_bit_resolution' not specified in detector characteristics."
            )

    @adc_bit_resolution.setter
    def adc_bit_resolution(self, value: int) -> None:
        """Set bit resolution of the Analog-Digital Converter."""
        self._adc_bit_resolution = value

    @property
    def adc_voltage_range(self) -> t.Tuple[float, float]:
        """Get voltage range of the Analog-Digital Converter."""
        if self._adc_voltage_range:
            return self._adc_voltage_range
        else:
            raise ValueError(
                "'adc_voltage_range' not specified in detector characteristics."
            )

    @adc_voltage_range.setter
    def adc_voltage_range(self, value: t.Tuple[float, float]) -> None:
        """Set voltage range of the Analog-Digital Converter."""
        self._adc_voltage_range = value

    @property
    def full_well_capacity(self) -> float:
        """Get Full well capacity."""
        if self._full_well_capacity:
            return self._full_well_capacity
        else:
            raise ValueError(
                "'full_well_capacity' not specified in detector characteristics."
            )

    @full_well_capacity.setter
    def full_well_capacity(self, value: float) -> None:
        """Set Full well capacity."""
        if value not in range(10000001):
            raise ValueError("'full_well_capacity' must be between 0 and 1e+7.")

        self._full_well_capacity = value

    @property
    def system_gain(self) -> float:
        """Get system gain."""
        return (
            self.quantum_efficiency
            * self.avalanche_gain
            * self.charge_to_volt_conversion
            * 2**self.adc_bit_resolution
        ) / (max(self.adc_voltage_range) - min(self.adc_voltage_range))

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes

    @staticmethod
    def bias_to_node_capacitance_saphira(bias: float) -> float:
        """Pixel integrating node capacitance in F.

        The below interpolates empirical published data, however note that
        Node C = Charge Gain / Voltage Gain
        So can be calculated by measuring V gain (varying PRV) and chg gain (PTC); see [2]

        Parameters
        ----------
        bias: float

        Returns
        -------
        output_capacitance: float
        """
        if bias < 1:
            raise ValueError(
                "Warning! Node capacitance calculation is inaccurate for bias voltages <1 V!"
            )

        # From [2] (Mk13 ME1000; data supplied by Leonardo):
        bias_list = [1, 1.5, 2.5, 3.5, 4.5, 6.5, 8.5, 10.5]
        capacitance = [46.5, 41.3, 37.3, 34.8, 33.2, 31.4, 30.7, 30.4]

        output_capacitance = float(np.interp(x=bias, xp=bias_list, fp=capacitance))

        return output_capacitance * 1.0e-15

    @staticmethod
    def bias_to_gain_saphira(bias: float) -> float:
        """Calculate gain from bias.

        The formula ignores the soft knee between the linear and unity gain ranges,
        but should be close enough. [2] (Mk13 ME1000)

        Parameters
        ----------
        bias: float

        Returns
        -------
        gain: float
        """

        gain = 2 ** ((bias - 2.65) / 2.17)

        if gain < 1.0:
            gain = 1.0  # Unity gain is lowest

        return gain

    @staticmethod
    def gain_to_bias_saphira(gain: float) -> float:
        """Calculate bias from gain.

        The formula ignores the soft knee between the linear and
        unity gain ranges, but should be close enough. [2] (Mk13 ME1000)

        Parameters
        ----------
        gain: float

        Returns
        -------
        bias: float
        """

        bias = (2.17 * math.log2(gain)) + 2.65

        return bias

    @staticmethod
    def detector_gain_saphira(capacitance: float, roic_gain: float) -> float:
        """Saphira detector gain.

        Parameters
        ----------
        capacitance: float
        roic_gain: float

        Returns
        -------
        float
        """
        # Late import to speedup start-up time
        import astropy.constants as const

        return roic_gain * (const.e.value / capacitance)

    def to_dict(self) -> t.Mapping:
        """Get the attributes of this instance as a `dict`."""
        if self._avalanche_gain and self._pixel_reset_voltage:
            dct = {
                "avalanche_gain": self._avalanche_gain,
                "common_voltage": None,
                "pixel_reset_voltage": self._pixel_reset_voltage,
            }
        elif self._avalanche_gain and self._common_voltage:
            dct = {
                "avalanche_gain": self._avalanche_gain,
                "common_voltage": self._common_voltage,
                "pixel_reset_voltage": None,
            }
        elif self._common_voltage and self._pixel_reset_voltage:
            dct = {
                "avalanche_gain": None,
                "common_voltage": self._common_voltage,
                "pixel_reset_voltage": self._pixel_reset_voltage,
            }
        else:
            raise NotImplementedError

        other_dct = {
            "quantum_efficiency": self._quantum_efficiency,
            "full_well_capacity": self._full_well_capacity,
            "adc_voltage_range": self._adc_voltage_range,
            "adc_bit_resolution": self._adc_bit_resolution,
            "roic_gain": self._roic_gain,
        }

        return {**dct, **other_dct}

    @classmethod
    def from_dict(cls, dct: t.Mapping):
        """Create a new instance from a `dict`."""
        if "adc_voltage_range" in dct:
            new_dct = dicttoolz.dissoc(dct, "adc_voltage_range")  # type: t.Mapping
            adc_voltage_range = dct["adc_voltage_range"]

            if adc_voltage_range is not None:
                adc_voltage_range = tuple(adc_voltage_range)

            return cls(adc_voltage_range=adc_voltage_range, **new_dct)
        else:
            return cls(**dct)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Detector class."""
import collections
import typing as t
from pathlib import Path

import numpy as np

from pyxel import backends
from pyxel.data_structure import (
    Charge,
    Image,
    Persistence,
    Photon,
    Pixel,
    Signal,
    SimplePersistence,
)
from pyxel.detectors import Environment
from pyxel.detectors.readout_properties import ReadoutProperties
from pyxel.util.memory import get_size, memory_usage_details

__all__ = ["Detector"]


# TODO: Add methods to save/load a `Detector` instance to the filesystem. See #329
class Detector:
    """The detector class."""

    def __init__(self, environment: Environment):
        self.environment = environment  # type: Environment

        self.header = collections.OrderedDict()  # type: t.Dict[str, object]

        self._photon = None  # type: t.Optional[Photon]
        self._charge = None  # type: t.Optional[Charge]
        self._pixel = None  # type: t.Optional[Pixel]
        self._signal = None  # type: t.Optional[Signal]
        self._image = None  # type: t.Optional[Image]

        # This will be the memory of the detector where trapped charges will be saved
        self._memory = dict()  # type: t.Dict
        self._persistence = (
            None
        )  # type: t.Optional[t.Union[Persistence, SimplePersistence]]

        self.input_image = None  # type: t.Optional[np.ndarray]
        self._output_dir = None  # type: t.Optional[Path]  # TODO: See #330

        self._readout_properties = None  # type: t.Optional["ReadoutProperties"]

        self._numbytes = get_size(self)

    def __eq__(self, other) -> bool:
        return (
            isinstance(other, Detector)
            and self._photon == other._photon
            and self._charge == other._charge
            and self._pixel == other._pixel
            and self._signal == other._signal
            and self._image == other._image
        )

    @property
    def geometry(self):
        """TBW."""
        raise NotImplementedError

    @property
    def characteristics(self):
        """TBW."""
        raise NotImplementedError

    @property
    def photon(self) -> Photon:
        """TBW."""
        if not self._photon:
            raise RuntimeError("Photon array is not initialized ! ")
        return self._photon

    @photon.setter
    def photon(self, obj: Photon) -> None:
        self._photon = obj

    @property
    def charge(self) -> Charge:
        """TBW."""
        if not self._charge:
            raise RuntimeError("'charge' not initialized.")

        return self._charge

    @property
    def pixel(self) -> Pixel:
        """TBW."""
        if not self._pixel:
            raise RuntimeError("'pixel' not initialized.")

        return self._pixel

    @property
    def signal(self) -> Signal:
        """TBW."""
        if not self._signal:
            raise RuntimeError("'signal' not initialized.")

        return self._signal

    @property
    def image(self) -> Image:
        """TBW."""
        if not self._image:
            raise RuntimeError("'image' not initialized.")

        return self._image

    def reset(self) -> None:
        """TBW."""
        self._photon = Photon(geo=self.geometry)
        self._charge = Charge(geo=self.geometry)
        self._pixel = Pixel(geo=self.geometry)
        self._signal = Signal(geo=self.geometry)
        self._image = Image(geo=self.geometry)

    def empty(self, empty_all: bool = True) -> None:
        """Empty the data in the detector.

        Returns
        -------
        None
        """
        if self._photon:
            self.photon.array *= 0
        if self._signal:
            self.signal.array *= 0
        if self._image:
            self.image.array *= 0
        if self._charge:
            self._charge.empty()
        if empty_all:
            if self._pixel:
                self.pixel.array *= 0

    # TODO: Set an `Output` object ? Is it really needed ? See #330
    def set_output_dir(self, path: t.Union[str, Path]) -> None:
        """Set output directory path."""
        self._output_dir = Path(path)

    # TODO: Set an `Output` object ? Is it really needed ? See #330
    @property
    def output_dir(self) -> Path:
        """Output directory path."""
        if self._output_dir is None:
            raise RuntimeError("'output_dir' is not initialized.")

        return self._output_dir

    def set_readout(
        self,
        num_steps: int,
        start_time: float,
        end_time: float,
        ndreadout: bool = False,
        times_linear: bool = True,
    ) -> None:
        """Set readout."""
        self._readout_properties = ReadoutProperties(
            num_steps=num_steps,
            start_time=start_time,
            end_time=end_time,
            ndreadout=ndreadout,
            times_linear=times_linear,
        )

    @property
    def time(self) -> float:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.time
        else:
            raise ValueError("No readout defined.")

    @time.setter
    def time(self, value: float) -> None:
        """TBW."""
        if self._readout_properties is not None:
            self._readout_properties.time = value
        else:
            raise ValueError("No readout defined.")

    @property
    def start_time(self) -> float:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.start_time
        else:
            raise ValueError("No readout defined.")

    @start_time.setter
    def start_time(self, value: float) -> None:
        """TBW."""
        if self._readout_properties is not None:
            self._readout_properties.start_time = value
        else:
            raise ValueError("No readout defined.")

    @property
    def absolute_time(self) -> float:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.absolute_time
        else:
            raise ValueError("No readout defined.")

    @property
    def time_step(self) -> float:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.time_step
        else:
            raise ValueError("No readout defined.")

    @time_step.setter
    def time_step(self, value: float) -> None:
        """TBW."""
        if self._readout_properties is not None:
            self._readout_properties.time_step = value
        else:
            raise ValueError("No readout defined.")

    @property
    def times_linear(self) -> bool:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.times_linear
        else:
            raise ValueError("No readout defined.")

    @property
    def num_steps(self) -> int:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.num_steps
        else:
            raise ValueError("No readout defined.")

    @property
    def pipeline_count(self) -> float:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.pipeline_count
        else:
            raise ValueError("No readout defined.")

    @pipeline_count.setter
    def pipeline_count(self, value: int) -> None:
        """TBW."""
        if self._readout_properties is not None:
            self._readout_properties.pipeline_count = value
        else:
            raise ValueError("No readout defined.")

    @property
    def read_out(self) -> bool:
        """TBW."""
        if self._readout_properties is not None:
            return self._readout_properties.read_out
        else:
            raise ValueError("No readout defined.")

    @read_out.setter
    def read_out(self, value: bool) -> None:
        """TBW."""
        if self._readout_properties is not None:
            self._readout_properties.read_out = value
        else:
            raise ValueError("No readout defined.")

    @property
    def is_dynamic(self) -> bool:
        """Return if detector is dynamic (time dependent) or not.

        By default it is not dynamic.
        """
        if self._readout_properties is not None:
            return True
        else:
            return False

    @property
    def non_destructive_readout(self) -> bool:
        """Return if detector readout mode is destructive or integrating.

        By default it is destructive (non-integrating).
        """
        if self._readout_properties is not None:
            return self._readout_properties.non_destructive
        else:
            raise ValueError("No sampling defined.")

    def has_persistence(self) -> bool:
        """TBW."""
        if self._persistence is not None:
            return True
        else:
            return False

    @property
    def persistence(self) -> t.Union[Persistence, SimplePersistence]:
        """TBW."""
        if self._persistence is not None:
            return self._persistence
        else:
            raise RuntimeError("'persistence' not initialized.")

    @persistence.setter
    def persistence(self, value: t.Union[Persistence, SimplePersistence]) -> None:
        """TBW."""
        if not isinstance(value, (Persistence, SimplePersistence)):
            raise TypeError(
                "Expecting Persistence or SimplePersistence type to set detector persistence."
            )
        self._persistence = value

    @property
    def numbytes(self) -> int:
        """Recursively calculates object size in bytes using Pympler library.

        Returns
        -------
        int
            Size of the object in bytes.
        """
        self._numbytes = get_size(self)
        return self._numbytes

    def memory_usage(
        self, print_result: bool = True, human_readable: bool = True
    ) -> dict:
        """TBW.

        Returns
        -------
        dict
            Dictionary of attribute memory usage
        """
        attributes = [
            "_photon",
            "_charge",
            "_pixel",
            "_signal",
            "_image",
            "material",
            "environment",
            "_geometry",
            "_characteristics",
        ]

        return memory_usage_details(
            self, attributes, print_result=print_result, human_readable=human_readable
        )

    # TODO: Move this to another place. See #241
    def to_hdf5(self, filename: t.Union[str, Path]) -> None:
        """Write the detector content to a :term:`HDF5` file.

        The HDF5 file has the following structure:

        .. code-block:: bash

            filename.h5  (4 objects, 3 attributes)
            │   ├── pyxel-version  1.0.0+161.g659eec86
            │   ├── type  CCD
            │   └── version  1
            ├── geometry  (5 objects)
            │   ├── col  (), int64
            │   ├── pixel_horz_size  (), float64
            │   ├── pixel_vert_size  (), float64
            │   ├── row  (), int64
            │   └── total_thickness  (), float64
            ├── environment  (1 object)
            │   └── temperature  (), float64
            ├── characteristics  (4 objects)
            │   ├── charge_to_volt_conversion  (), float64
            │   ├── full_well_capacity  (), int64
            │   ├── pre_amplification  (), float64
            │   └── quantum_efficiency  (), float64
            └── data  (5 objects)
                ├── charge  (2 objects, 2 attributes)
                │   ├── name  Charge
                │   ├── unit  electron
                │   ├── array  (100, 120), float64
                │   └── frame  (13 objects, 1 attribute)
                │       ├── type  DataFrame
                │       ├── charge  (0,), float64
                │       ├── energy  (0,), float64
                │       ├── init_energy  (0,), float64
                │       ├── init_pos_hor  (0,), float64
                │       ├── init_pos_ver  (0,), float64
                │       ├── init_pos_z  (0,), float64
                │       ├── number  (0,), float64
                │       ├── position_hor  (0,), float64
                │       ├── position_ver  (0,), float64
                │       ├── position_z  (0,), float64
                │       ├── velocity_hor  (0,), float64
                │       ├── velocity_ver  (0,), float64
                │       └── velocity_z  (0,), float64
                ├── image  (100, 120), uint64
                │   └── name  Image
                ├── photon  (100, 120), float64
                ├── pixel  (100, 120), float64
                └── signal  (100, 120), float64

        Parameters
        ----------
        filename : str or Path

        Notes
        -----
        You can find more information in the 'how-to' guide section.

        Examples
        --------
        >>> from pyxel.detectors import CCD
        >>> detector = CCD(...)

        >>> detector.to_hdf5("ccd.h5")
        """
        dct = self.to_dict()  # type: t.Mapping
        backends.to_hdf5(filename=filename, dct=dct)

    @classmethod
    def from_hdf5(cls, filename: t.Union[str, Path]) -> "Detector":
        """Load a detector object from a :term:`HDF5` file.

        Parameters
        ----------
        filename : str or Path

        Examples
        --------
        >>> detector = Detector.from_hdf5("ccd.h5")
        >>> detector
        CCD(...)
        """
        dct = backends.from_hdf5(filename)  # type: t.Mapping[str, t.Any]

        obj = cls.from_dict(dct)  # type: Detector
        return obj

    def to_dict(self) -> t.Mapping:
        """Convert a `Detector` to a `dict`."""
        raise NotImplementedError

    # TODO: Replace `-> 'Detector'` by `t.Union[CCD, CMOS, MKID]`
    @classmethod
    def from_dict(cls, dct: t.Mapping) -> "Detector":
        """Create a new instance of a `Detector` from a `dict`."""
        # TODO: This is a simplistic implementation. Improve this.
        if dct["type"] == "CCD":
            from pyxel.detectors import CCD  # Imported here to avoid circular import

            return CCD.from_dict(dct)

        elif dct["type"] == "CMOS":
            from pyxel.detectors import CMOS

            return CMOS.from_dict(dct)

        elif dct["type"] == "MKID":
            from pyxel.detectors import MKID

            return MKID.from_dict(dct)

        elif dct["type"] == "APD":
            from pyxel.detectors import APD

            return APD.from_dict(dct)

        else:
            raise NotImplementedError(f"Unknown type: {dct['type']!r}")



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.
#
#
"""Avalanche gain model."""
from pyxel.detectors import APD


def apd_gain(detector: APD) -> None:
    """Apply APD gain.

    Parameters
    ----------
    detector: APD
        Pyxel APD detector object.
    """

    if not isinstance(detector, APD):
        raise TypeError("Expecting a 'APD' detector object.")

    if detector.charge.frame_empty():

        array_copy = detector.charge.array.copy()
        detector.charge.empty()
        array_copy *= detector.characteristics.avalanche_gain
        detector.charge.add_charge_array(array=array_copy)

    else:

        frame_copy = detector.charge.frame.copy()
        detector.charge.empty()
        frame_copy.number *= detector.characteristics.avalanche_gain
        detector.charge.add_charge_dataframe(new_charges=frame_copy)



#  Copyright (c) European Space Agency, 2017, 2018, 2019, 2020, 2021, 2022.
#
#  This file is subject to the terms and conditions defined in file 'LICENCE.txt', which
#  is part of this Pyxel package. No part of the package, including
#  this file, may be copied, modified, propagated, or distributed except according to
#  the terms contained in the file ‘LICENCE.txt’.

"""Fitness functions for model fitting."""

import numba
import numpy as np


@numba.njit
def sum_of_abs_residuals(
    simulated: np.ndarray,
    target: np.ndarray,
    weighting: np.ndarray,
) -> float:
    """TBW.

    Parameters
    ----------
    simulated
    target
    weighting

    Returns
    -------
    array
        TBW.
    """
    diff = target - simulated
    diff *= weighting

    result = float(np.sum(np.abs(diff)))
    return result


@numba.njit
def sum_of_squared_residuals(
    simulated: np.ndarray,
    target: np.ndarray,
    weighting: np.ndarray,
) -> float:
    """TBW.

    Parameters
    ----------
    simulated
    target
    weighting

    Returns
    -------
    array
        TBW.
    """
    diff = target - simulated
    diff_square = diff * diff
    diff_square *= weighting

    result = float(np.sum(diff_square))
    return result


@numba.njit
def reduced_chi_squared(
    simulated: np.ndarray,
    target: np.ndarray,
    weighting: np.ndarray,
    free_parameters: int,
) -> float:
    r"""Compute the reduced chi-square error statistic.

    Notes
    -----
    You can find more information at this link
    https://en.wikipedia.org/wiki/Goodness_of_fit

    Parameters
    ----------
    simulated
    target
    weighting
    free_parameters : int
        Number of free parameters in the model

    Returns
    -------
    float
        The reduced :math:`\chi^{2}`.
    """
    diff = target - simulated
    deviation2 = np.square(diff / weighting)

    degrees_of_freedom = target.size - free_parameters
    reduced_chi2 = float(np.sum(deviation2)) / degrees_of_freedom

    return reduced_chi2
